@@ -19,7 +19,7 @@ from xml.etree import ElementTree as ET
 import thumbhash
 from PIL import Image, ImageOps
 from slixmpp import JID, Message
-from slixmpp.exceptions import IqError
+from slixmpp.exceptions import IqError, IqTimeout
 from slixmpp.plugins.xep_0363 import FileUploadError
 from slixmpp.plugins.xep_0447.stanza import StatelessFileSharing
 
@@ -65,7 +65,7 @@ class AttachmentMixin(TextMessageMixin):
                 ifrom=config.UPLOAD_REQUESTER or self.xmpp.boundjid,
                 domain=JID(domain),
             )
-        except (FileUploadError, IqError) as e:
+        except (FileUploadError, IqError, IqTimeout) as e:
             warnings.warn(f"Something is wrong with the upload service: {e!r}")
             return None
         finally:
@@ -374,9 +374,20 @@ class AttachmentMixin(TextMessageMixin):
                 if xmpp_id == original_xmpp_id:
                     continue
                 self.retract(xmpp_id, thread)
+
+        if reply_to is not None and reply_to.body:
+            # We cannot have a "quote fallback" for attachments since most (all?)
+            # XMPP clients will only treat a message as an attachment if the
+            # body is the URL and nothing else.
+            reply_to_for_attachment: MessageReference | None = MessageReference(
+                reply_to.legacy_id, reply_to.author
+            )
+        else:
+            reply_to_for_attachment = reply_to
+
         msg = self._make_message(
             when=when,
-            reply_to=reply_to,
+            reply_to=reply_to_for_attachment,
             carbon=carbon,
             mto=mto,
             thread=thread,
